# coding=utf-8
"""This module is an adapter for the NRAO Solr index.

The module provides mechanisms for passing parameters to the solr index and return
a results set in various forms; paginated execution blocks, results grouped by
project, calibration files, etc.

"""
from __future__ import print_function
import logging
import json
from string import Template
from pyramid.view import view_config
from pycapo import CapoConfig
import requests
from services.__init__ import base_repr
from services.solr.__init__ import def_name
from services.solr.query_builder import ArchiveSolrParameters, construct_execution_block_query, \
    construct_images_query, append_eb_column_filters_query

_LOG = logging.getLogger(__name__)

_APP_X_WWW_FORM_URLENCODED = 'application/x-www-form-urlencoded'
_SOLR_QUERY_FAILED_WITH_CODE = Template('solr query failed with code: $error_msg')

_EXCEPTION_MSG = Template('Exception caught in $c.$d: $code')
_REQUEST_PROCESS_SUCCESSFULLY = 'Request processed successfully'
_NO_XXX_FOUND = Template('No $name found')


class SolrException(Exception):
    r"""An exception to make our internal exceptions less 'generic' for code inspection"""


class SolrNamedViews:
    r"""This class is a container for the archive service layer endpoints that interface
    directly with the archive solr indexes"""

    def __init__(self, request):
        r"""Constructor for the class"""
        super(SolrNamedViews, self).__init__()
        self.profile = request.session['capo_profile']
        self.request = request
        self.solr = SolrInterface(self.profile)

    def __repr__(self):
        base_repr(self)

    @view_config(route_name='get_eb_project_view', renderer='jsonp')
    def get_eb_project_view(self):
        r"""Return a paginated list of projects from solr"""
        project_dict = None
        try:
            project_dict = self.solr.get_paged_eb_projects(**self.request.params)

            response_msg = ResponseMsg(True, ResponseMsg.SUCCESS, _REQUEST_PROCESS_SUCCESSFULLY)
        except SolrException as ex_msg:
            _LOG.info(ex_msg)
            response_msg = ResponseMsg(False, ResponseMsg.HARD_ERROR, str(ex_msg))
        return json.dumps(dict(project_dict=project_dict,
                               n_results=project_dict['n_projects'] if project_dict else 0,
                               data=project_dict['projects'] if project_dict else None,
                               response_msg=response_msg.as_dict()))

    @view_config(route_name='get_solr_facets', renderer='jsonp')
    def get_solr_facets(self):
        r"""Get two lists of facets from solr, these are small lists of {field: [unique value]}
        that can be used by a client for such things as pull-down menus, since they represent
        values in solr for those fields, they are guaranteed to be in the index.
        """
        facet_list = None
        img_facet_list = None
        try:
            facet_list = self.solr.get_eb_core_facets()
            img_facet_list = self.solr.get_img_solr_facets()
            if not facet_list:
                response_msg = ResponseMsg(False, ResponseMsg.HARD_ERROR,
                                           '{0} error: facet list empty'.format(def_name()))
            else:
                response_msg = ResponseMsg(True, ResponseMsg.SUCCESS, _REQUEST_PROCESS_SUCCESSFULLY)
        except SolrException as ex_msg:
            _LOG.info(ex_msg)
            response_msg = ResponseMsg(False, ResponseMsg.HARD_ERROR, str(ex_msg))
        return json.dumps(dict(facet_list=facet_list,
                               img_facet_list=img_facet_list,
                               response_msg=response_msg.as_dict()))

    @view_config(route_name='get_paged_exec_blocks', renderer='jsonp')
    def get_paged_exec_blocks(self):
        r"""Return a paginated list of exec_block documents from solr, in the form of a json
        blob."""
        eb_list = None
        eb_total = 0
        try:
            eb_list, eb_total = self.solr.get_paged_exec_blocks(**self.request.params)
            response_msg = ResponseMsg(True, ResponseMsg.SUCCESS, _REQUEST_PROCESS_SUCCESSFULLY)
        except SolrException as ex_msg:
            _LOG.info(ex_msg)
            response_msg = ResponseMsg(False, ResponseMsg.HARD_ERROR, str(ex_msg))
        return json.dumps(dict(eb_list=eb_list,
                               eb_total=eb_total,
                               n_results=eb_total,
                               data=eb_list,
                               response_msg=response_msg.as_dict()))

    @view_config(route_name='get_paged_alma_mouses', renderer='jsonp')
    def get_paged_alma_mouses(self):
        r"""Return a paginated list of ALMA mouses from solr, in the form of a json blob."""
        mous_list = None
        mous_count = 0
        try:
            mous_list, mous_count = self.solr.get_paged_mous_groups(**self.request.params)
            response_msg = ResponseMsg(True, ResponseMsg.SUCCESS, _REQUEST_PROCESS_SUCCESSFULLY)
        except SolrException as ex_msg:
            _LOG.info(ex_msg)
            response_msg = ResponseMsg(False, ResponseMsg.HARD_ERROR, str(ex_msg))
        return json.dumps(dict(mouses=mous_list,
                               data=mous_list,
                               n_results=mous_count,
                               response_msg=response_msg.as_dict()))

    @view_config(route_name='get_full_exec_block_details', renderer='jsonp')
    def get_full_exec_block_details(self):
        r"""Return the full details of an exec_block document from solr, including it's subscans,
        in the form of a json blob."""
        n_results = 0
        curr_eb = None
        try:
            curr_eb = self.solr.get_full_exec_block_details(**self.request.params)
            if curr_eb:
                n_results = 1
                response_msg = ResponseMsg(True,
                                           ResponseMsg.SUCCESS, _REQUEST_PROCESS_SUCCESSFULLY)
            else:
                response_msg = ResponseMsg(False,
                                           ResponseMsg.HARD_ERROR, 'Exec Block ID not found.')
        except SolrException as ex_msg:
            _LOG.info(ex_msg)
            response_msg = ResponseMsg(False, ResponseMsg.HARD_ERROR, str(ex_msg))
        return json.dumps(dict(curr_eb=curr_eb,
                               n_results=n_results,
                               data=curr_eb,
                               response_msg=response_msg.as_dict()))

    @view_config(route_name='get_paged_images', renderer='jsonp')
    def get_paged_images(self):
        r"""Return a paginated list of images from the solr index, in the form of a json blob."""
        img_list = None
        img_total = 0
        try:
            img_list, img_total = self.solr.get_paged_images(**self.request.params)
            response_msg = ResponseMsg(True, ResponseMsg.SUCCESS, _REQUEST_PROCESS_SUCCESSFULLY)
        except SolrException as ex:
            ex_msg = _EXCEPTION_MSG.substitute(c=__name__, d=def_name(), code=ex)
            _LOG.info(ex_msg)
            response_msg = ResponseMsg(False, ResponseMsg.HARD_ERROR, ex_msg)
        return json.dumps(dict(img_list=img_list,
                               img_total=img_total,
                               data=img_list,
                               n_results=img_total,
                               response_msg=response_msg.as_dict()))


class SolrInterface:
    """
        The main interface object for the NRAO solr index.  Each function
        defines a specific type of query, returning a dictionary of the resulting
        solr document fields.
    """

    def __init__(self, profile):
        r"""Constructor for the class"""
        try:
            capo_config = CapoConfig(profile=profile)
        except ValueError:
            raise Exception('No capo profile provided.')

        try:
            self.solr_url = capo_config.get('edu.nrao.archive.archiveIface.solrBaseUrl')
            self.thumbnail_url_base = capo_config.get(
                'edu.nrao.archive.archiveIface.thumbnailURLPrefix')
            self.solr_eb_core_url = self.solr_url + '/execblocks/query'
            self.solr_cal_core_url = self.solr_url + '/cals/query'
            self.solr_img_core_url = self.solr_url + '/images/query'
        except KeyError as key_ex:
            raise Exception('Capo key error: {0}'.format(key_ex))

    @staticmethod
    def get_simple_data_blob(query, start, rows, sort):
        r"""A static method to bundle the most basic solr POST.data"""
        return {
            'wt': 'json',
            'q': query,
            'start': start,
            'rows': rows,
            'sort': sort
        }

    @staticmethod
    def get_solr_query_data_block(query_params, q_obj):
        r"""Get the basic solr query data block, to include the spacial parameters"""
        data = SolrInterface.get_simple_data_blob(q_obj['q'],
                                                  query_params.start,
                                                  query_params.rows,
                                                  query_params.sort)
        if q_obj['geofilt']:
            data['fq'] = str(q_obj['geofilt']['fq'])
            data['pt'] = str(q_obj['geofilt']['pt'])
            data['d'] = str(q_obj['geofilt']['d'])
        return data

    def query_eb_core(self, query_params, q_obj, group_on=None):
        r"""Return a query of the execblocks solr index, optionally grouping."""
        data = SolrInterface.get_solr_query_data_block(query_params, q_obj)
        if group_on is not None:
            data['group'] = 'true'
            data['group.field'] = group_on
            data['group.ngroups'] = 'true'
            data['group.sort'] = query_params.sort

        return requests.post(self.solr_eb_core_url,
                             headers={'content-type': _APP_X_WWW_FORM_URLENCODED},
                             timeout=5,
                             data=data)

    @staticmethod
    def mous_gous_sous_dict(source_result):
        r"""A static method to the mous, gous, and sous IDs from a solr result"""
        ret_dict = dict()
        ret_dict['mousId'] = source_result.get('mousId', None)
        ret_dict['gousId'] = source_result.get('gousId', None)
        ret_dict['sousId'] = source_result.get('sousId', None)
        return ret_dict

    @staticmethod
    def dq_notes(solr_result):
        r"""A static method to bundle the data quality notes fields from a solr result"""
        ret_dict = dict()
        ret_dict['comments'] = solr_result.get('comments', None)
        ret_dict['problem_severity'] = solr_result.get('problem_severity', None)
        ret_dict['problem_type'] = solr_result.get('problem_type', None)
        return ret_dict

    @staticmethod
    def build_basic_eb_dict(solr_result):
        r"""A method to bundle an exec_block document for a solr result"""
        ret_dict = dict()
        ret_dict['id'] = solr_result.get('id')
        ret_dict['eb_id'] = int(solr_result.get('eb_id'))
        ret_dict['sci_prod_locator'] = solr_result.get('sci_prod_locator')
        ret_dict['obs_id'] = solr_result.get('obs_id', _NO_XXX_FOUND.substitute(name='obs id'))
        ret_dict['schedBlockName'] = solr_result.get('schedBlockName', None)
        ret_dict['filegroup_id'] = solr_result.get('filegroup_id', None)
        ret_dict['legacy_id'] = solr_result.get('legacy_id', None)
        ret_dict['project_code'] = solr_result.get('project_code',
                                                   _NO_XXX_FOUND.substitute(name='project code'))
        ret_dict['title'] = solr_result.get('title', _NO_XXX_FOUND.substitute(name='title'))
        ret_dict['abstract'] = solr_result.get('abstract',
                                               _NO_XXX_FOUND.substitute(name='abstract'))
        ret_dict['cal_status'] = solr_result.get('cal_status', 'Unknown')
        ret_dict['pi'] = solr_result.get('pi', 'No PI found')
        ret_dict['obs_start'] = solr_result.get('obs_start', None)
        ret_dict['obs_stop'] = solr_result.get('obs_stop', None)
        ret_dict['obs_release_date'] = solr_result.get('obs_release_date', None)
        ret_dict['proj_start'] = solr_result.get('proj_start', None)
        ret_dict['proj_stop'] = solr_result.get('proj_stop', None)
        ret_dict['instrument_name'] = \
            solr_result.get('instrument_name', _NO_XXX_FOUND.substitute(name='instrument name'))
        ret_dict['data_rights'] = solr_result.get('data_rights', 'LOCKED')
        ret_dict['calib_level'] = int(solr_result.get('calib_level')) \
            if solr_result.get('calib_level', None) else None
        ret_dict['vla_configuration'] = solr_result.get('vla_configuration') \
            if solr_result.get('vla_configuration', None) and \
            not solr_result.get('vla_configuration') == 'Unknown' else None
        ret_dict['dataproduct_type'] = solr_result.get('dataproduct_type', None)
        ret_dict['access_estsize'] = int(solr_result.get('access_estsize', 0))
        ret_dict['num_scans'] = solr_result.get('num_scans', 0)
        ret_dict['obs_band'] = solr_result.get('obs_band', [])
        ret_dict['intents'] = solr_result.get('eb_scan_intents', [])
        ret_dict['coIs'] = solr_result.get('coIs', None)
        ret_dict.update(SolrInterface.dq_notes(solr_result))
        ret_dict.update(SolrInterface.mous_gous_sous_dict(solr_result))
        return ret_dict

    @staticmethod
    def build_basic_cal_dict(solr_result):
        r"""A method to build a cal document from a solr result"""
        ret_dict = dict()
        ret_dict['id'] = solr_result.get('id')
        ret_dict['exec_id'] = int(solr_result.get('exec_id')) \
            if solr_result.get('exec_id', None) else None
        ret_dict['sci_prod_locator'] = solr_result.get('sci_prod_locator')
        ret_dict['obs_id'] = solr_result.get('obs_id', _NO_XXX_FOUND.substitute(name='obs id'))
        ret_dict['project_code'] = solr_result.get('project_code',
                                                   _NO_XXX_FOUND.substitute(name='project code'))
        ret_dict['file_name'] = solr_result.get('file_name', None)
        ret_dict.update(SolrInterface.dq_notes(solr_result))
        ret_dict.update(SolrInterface.mous_gous_sous_dict(solr_result))
        return ret_dict

    def build_basic_img_dict(self, solr_result):
        r"""A method to bundle an image document from a solr result"""
        ret_dict = dict()

        ret_dict['id'] = solr_result.get('id')
        ret_dict['sci_prod_locator'] = solr_result.get('sci_prod_locator')
        ret_dict['project_code'] = solr_result.get('project_code', None)
        ret_dict['file_name'] = solr_result.get('file_name')
        ret_dict['thumbnail'] = self.thumbnail_url_base + '/' + solr_result.get('thumbnail') \
            if solr_result.get('thumbnail', None) else None
        ret_dict['access_estsize'] = int(solr_result.get('access_estsize')) \
            if solr_result.get('access_estsize', None) else None
        ret_dict['sp_resolution'] = float(solr_result.get('sp_resolution')) \
            if solr_result.get('sp_resolution', None) else None
        ret_dict['beam_axis_ratio'] = float(solr_result.get('beam_axis_ratio')) \
            if solr_result.get('beam_axis_ratio', None) else None
        ret_dict['instrument_name'] = solr_result.get('instrument_name', None)
        ret_dict['band_code'] = solr_result.get('band_code', None)
        ret_dict['target_pos_equatorial'] = solr_result.get('target_pos_equatorial', None)[0]
        ret_dict['target_pos_ecliptic'] = solr_result.get('target_pos_ecliptic', None)[0]
        ret_dict['target_pos_galactic'] = solr_result.get('target_pos_galactic', None)[0]
        ret_dict['min_frequency'] = solr_result.get('min_frequency', None)
        ret_dict['max_frequency'] = solr_result.get('max_frequency', None)
        ret_dict['polarization'] = solr_result.get('polarization', None)
        ret_dict['min_intensity'] = solr_result.get('min_intensity', None)
        ret_dict['max_intensity'] = solr_result.get('max_intensity', None)
        ret_dict['rms_noise'] = solr_result.get('rms_noise', None)
        ret_dict.update(SolrInterface.dq_notes(solr_result))
        return ret_dict

    def get_paged_eb_projects(self, **params_in):
        r"""Get a paginated list of exec_block results by querying the solr index using the
        given parameters, grouped as projects"""
        try:
            query_params = ArchiveSolrParameters(**params_in)
            q_obj = construct_execution_block_query(query_params)
            solr_result = self.query_eb_core(query_params, q_obj, group_on='project_code')

            if solr_result.status_code == 200:
                p_results = solr_result.json()
                grouped_results = p_results['grouped']['project_code']

                project_dict = dict()
                project_dict['n_projects'] = grouped_results['ngroups']
                proj_list = []
                for project_group in grouped_results['groups']:
                    deref_proj = project_group['doclist']['docs'][0]
                    curr_proj = self.build_basic_eb_dict(deref_proj)

                    curr_proj['n_execution_blocks'] = int(project_group['doclist']['numFound'])

                    proj_list.append(curr_proj)
                project_dict['projects'] = proj_list

                return project_dict

            # The server could be down 404 (json decode will fail)
            # or we could have a malformed query 501(?) (with a json serializable response)
            raise Exception(
                _SOLR_QUERY_FAILED_WITH_CODE.substitute(error_msg=solr_result.status_code))

        except Exception as any_excep:
            ex = _EXCEPTION_MSG.substitute(c=__name__, d=def_name(), code=any_excep)
            raise SolrException(ex)

    def get_paged_mous_groups(self, **params_in):
        r"""Get a paginated list of mouses by querying the solr index using the
        given parameters"""
        try:
            query_params = ArchiveSolrParameters(**params_in)
            q_obj = construct_execution_block_query(query_params)
            solr_result = self.query_eb_core(query_params, q_obj, group_on='mousId')

            if solr_result.status_code == 200:
                mous_group = dict()
                mous_list = []
                for group in solr_result.json()['grouped']['mousId']['groups']:
                    base_exec_block = group['doclist']['docs'][0]

                    curr_eb = SolrInterface.build_basic_eb_dict(base_exec_block)
                    curr_eb['eb_count'] = group['doclist']['numFound']

                    mous_group[base_exec_block.get('mousId')] = curr_eb
                    mous_list.append(base_exec_block.get('mousId'))

                cal_dicts = self.build_cal_dict_by_id_from_cal_list('mousId',
                                                                    self.get_paged_eb_calibrations(
                                                                        mous_list, alma=True))

                for exec_block in mous_group:
                    mous_size = self.get_mous_size(exec_block,
                                                   mous_group[exec_block]['eb_count'], q_obj['q'])
                    mous_group[exec_block]['access_estsize'] = mous_size

                    if exec_block in cal_dicts:
                        mous_group[exec_block]['cals'] = cal_dicts[exec_block]

                return mous_group, solr_result.json()['grouped']['mousId']['ngroups']

            # The server could be down 404 (json decode will fail)
            # or we could have a malformed query 501(?) (with a json serializable response)
            raise Exception(
                _SOLR_QUERY_FAILED_WITH_CODE.substitute(error_msg=solr_result.status_code))

        except Exception as any_except:
            ex = _EXCEPTION_MSG.substitute(c=__name__, d=def_name(), code=any_except)
            raise SolrException(ex)

    @staticmethod
    def get_facet_data_blob(fields):
        r"""Get a POST.data block for a solr facet query of the fields passed in"""
        data = SolrInterface.get_simple_data_blob('*:*', 0, 1, None)
        data['facet'] = 'on'
        data['facet.method'] = 'enum'
        data['facet.field'] = fields
        return data

    @staticmethod
    def parse_eb_facet_fields(facet_field_list):
        r"""Parse the facet fields from a solr facet query into the form
        [{field:{name:  facet,
                count: facet count},
                ...]
        """
        ret_dict = dict()
        for field_name in facet_field_list:
            curr_facets = {}
            # For each 'facet_field' we get an array, like:
            # [facet_1, facet_1_count, facet_2, facet_2_count...]
            for i in range(0, len(facet_field_list[field_name]), 2):
                # We'll iterate over our facet list in steps of two, for facet and count
                facet = facet_field_list[field_name][i]
                facet_count = facet_field_list[field_name][i + 1]

                if field_name == 'vla_configuration' and facet == 'Unknown':
                    # SSA-4818: For the moment, we don't want to display the
                    # 'Unknown' array configurations, which seem to be predominantly ALMA.
                    pass
                else:
                    # all else, just add the key and count
                    curr_facets[facet] = facet_count
            if field_name == 'instrument_name':
                curr_facets = {k.upper(): curr_facets[k] for k in sorted(curr_facets.keys())}
            sorted_curr_facets = {k: curr_facets[k] for k in sorted(curr_facets.keys())}
            ret_dict[field_name] = [dict(name=thisKey, count=thisValue) for thisKey, thisValue in
                                    sorted_curr_facets.items()]
        return ret_dict

    def get_eb_core_facets(self):
        r"""Get the facts for the execblocks solr index core"""
        data = SolrInterface.get_facet_data_blob(fields=['instrument_name', 'vla_configuration',
                                                         'obs_band', 'polarizations'])
        solr_result = requests.post(self.solr_eb_core_url,
                                    headers={'content-type': _APP_X_WWW_FORM_URLENCODED},
                                    timeout=5,
                                    data=data)
        if solr_result.status_code == 200:
            return SolrInterface.parse_eb_facet_fields(
                solr_result.json()['facet_counts']['facet_fields'])

        # The server could be down 404 (json decode will fail)
        # or we could have a malformed query 501(?) (with a json serializable response)
        raise SolrException(
            _SOLR_QUERY_FAILED_WITH_CODE.substitute(error_msg=solr_result.status_code))

    def get_img_solr_facets(self):
        r"""Get the facets for the images solr index core"""
        data = SolrInterface.get_facet_data_blob(fields=['polarization'])
        solr_result = requests.post(self.solr_img_core_url,
                                    headers={'content-type': _APP_X_WWW_FORM_URLENCODED},
                                    timeout=5,
                                    data=data)
        if solr_result.status_code == 200:
            facet_field_list = solr_result.json()['facet_counts']['facet_fields']
            ret_dict = dict()
            for key in facet_field_list:
                for i in range(0, len(facet_field_list[key]), 2):
                    # We'll iterate over our facet list in steps of two, for facet and count
                    facet = facet_field_list[key][i]
                    facet_count = facet_field_list[key][i + 1]
                    ret_dict[key] = [dict(name=facet, count=facet_count)]
            return ret_dict

        # The server could be down 404 (json decode will fail)
        # or we could have a malformed query 501(?) (with a json serializable response)
        raise SolrException(
            _SOLR_QUERY_FAILED_WITH_CODE.substitute(error_msg=solr_result.status_code))

    @staticmethod
    def build_cal_dict_by_id_from_cal_list(id_type, cal_list_in):
        r"""A static method to parse over a cal file list and group them into a dict of
        {'id_type': [cals]}"""
        cal_dict = dict()
        for cals in cal_list_in:
            if cals[id_type] not in cal_dict:
                cal_dict[cals[id_type]] = [cals]
            else:
                cal_dict[cals[id_type]].append(cals)
        return cal_dict

    def attach_cals_to_ebs(self, eb_list):
        r"""A method that modifies the eb_list passed in, attaching the cals that match the either
        the eb_list.eb_id or the eb_list.mousId"""
        mous_id_list = []
        eb_id_list = []
        for curr_eb in eb_list:
            if curr_eb.get('instrument_name') == 'ALMA':
                mous_id_list.append(curr_eb['mousId'])
            else:
                eb_id_list.append(curr_eb['eb_id'])

        # Rather than ping the solr index for each eb, as we're constructing our list of
        # dictionaries, we've constructed a list of eb ids for these paginated ebs, that we can
        # get from solr in one big blog.  Then we can iterate over our list again to match up
        # the cal to the exec block.
        paged_eb_cals = self.get_paged_eb_calibrations(eb_id_list)
        paged_mous_cals = self.get_paged_eb_calibrations(mous_id_list, alma=True)

        cal_dicts = SolrInterface.build_cal_dict_by_id_from_cal_list('exec_id', paged_eb_cals)
        cal_dicts.update(
            SolrInterface.build_cal_dict_by_id_from_cal_list('mousId', paged_mous_cals))

        for curr_eb in eb_list:
            if curr_eb['eb_id'] in cal_dicts:
                curr_eb['cals'] = cal_dicts[curr_eb['eb_id']]
            if curr_eb['mousId'] in cal_dicts:
                curr_eb['cals'] = cal_dicts[curr_eb['mousId']]
        return eb_list

    def get_paged_exec_blocks(self, **params_in):
        r"""Get a paginated list of exec_block results by querying the solr index using the
        given parameters"""
        try:
            query_params = ArchiveSolrParameters(**params_in)
            q_obj = construct_execution_block_query(query_params)
            if query_params.column_filters is not None:
                q_obj['q'] = append_eb_column_filters_query(q_obj['q'], query_params)
            solr_results = self.query_eb_core(query_params, q_obj)

            if solr_results.status_code == 200:
                eb_list = []
                eb_rspns = solr_results.json()['response']
                eb_total = eb_rspns['numFound']
                for eb_doc in eb_rspns['docs']:
                    eb_list.append(SolrInterface.build_basic_eb_dict(eb_doc))
                    eb_list = self.attach_cals_to_ebs(eb_list)

                return eb_list, eb_total

            raise Exception(
                _SOLR_QUERY_FAILED_WITH_CODE.substitute(error_msg=solr_results.status_code))

        except Exception as any_except:
            ex = _EXCEPTION_MSG.substitute(c=__name__, d=def_name(), code=any_except)
            raise SolrException(ex)

    def get_full_exec_block_details(self, **params_in):
        r"""Get a the deep details of an exec_block by querying the solr index using the
        solr_id"""
        try:
            query_params = ArchiveSolrParameters(**params_in)
            data = SolrInterface.get_simple_data_blob('id:"' + query_params.solr_id + '"',
                                                      0, 1, None)
            solr_rspns = requests.post(self.solr_eb_core_url,
                                       headers={'content-type': _APP_X_WWW_FORM_URLENCODED},
                                       timeout=5,
                                       data=data)
            if solr_rspns.status_code == 200:
                curr_eb = SolrInterface.build_basic_eb_dict(
                    solr_rspns.json()['response']['docs'][0])
            else:
                # The server could be down 404 (json decode will fail)
                # or we could have a malformed query 501(?) (with a json serializable response)
                raise Exception(
                    _SOLR_QUERY_FAILED_WITH_CODE.substitute(error_msg=solr_rspns.status_code))
            data = SolrInterface.get_simple_data_blob(
                '_root_:"' + query_params.solr_id + '" AND _docType_:subscan',
                0, 1, None)
            peek_subscan_size = requests.post(self.solr_eb_core_url,
                                              headers={'content-type': _APP_X_WWW_FORM_URLENCODED},
                                              timeout=5,
                                              data=data)

            if peek_subscan_size.status_code == 200:
                data = SolrInterface.get_simple_data_blob(
                    '_root_:"' + query_params.solr_id + '" AND _docType_:subscan',
                    0, peek_subscan_size.json()['response']['numFound'], None)
                solr_rspns = requests.post(self.solr_eb_core_url,
                                           headers={'content-type': _APP_X_WWW_FORM_URLENCODED},
                                           timeout=5,
                                           data=data)
                eb_subscans = solr_rspns.json()['response']['docs']
            else:
                # The server could be down 404 (json decode will fail)
                # or we could have a malformed query 501(?) (with a json serializable response)
                raise Exception(_SOLR_QUERY_FAILED_WITH_CODE.substitute(
                    error_msg=peek_subscan_size.status_code))

            scans = dict()
            for subscan in eb_subscans:
                if 'scan_id' in subscan:
                    curr_target = subscan.get('scan_id')

                    if curr_target not in scans:
                        curr_subscan = dict()
                        curr_subscan['id'] = subscan.get('scan_id')
                        curr_subscan['eb_id'] = subscan.get('eb_id')
                        curr_subscan['target_name'] = subscan.get('target_name')[0]
                        curr_subscan['scan_intent'] = subscan.get('scan_intent', [])
                        curr_subscan['t_resolution'] = float(subscan.get('t_resolution'))
                        curr_subscan['target_pos_equatorial'] = \
                            subscan.get('sbsc_target_pos_equatorial')[0]
                        curr_subscan['target_pos_ecliptic'] = \
                            subscan.get('sbsc_target_pos_ecliptic')[0]
                        curr_subscan['target_pos_galactic'] = \
                            subscan.get('sbsc_target_pos_galactic')[0]
                        curr_subscan['max_frequency'] = float(subscan.get('max_frequency'))
                        curr_subscan['min_frequency'] = float(subscan.get('min_frequency'))
                        curr_subscan['polarizations'] = subscan.get('polarizations', [])
                        scans[curr_target] = curr_subscan

            # Now that we have a dictionary of unique subscans, we can further reduce the
            # memory footprint of the data going out to the front end by converting our
            # dictionary to a simple array, without the long key.
            curr_eb['subscans'] = [scans[s] for s in scans]

            if curr_eb.get('instrument_name') == 'ALMA':
                eb_cals = self.get_exec_block_calibrations(curr_eb.get('mousId'), alma=True)
            else:
                eb_cals = self.get_exec_block_calibrations(curr_eb.get('eb_id'))
            curr_eb['cals'] = eb_cals

            return curr_eb

        except Exception as any_except:
            ex = _EXCEPTION_MSG.substitute(c=__name__, d=def_name(), code=any_except)
            raise SolrException(ex)

    def query_cals(self, query):
        r"""Get the cals associated with the query passed, which could be for a single or
        multiple eb_ids or mousIds"""
        ret_cal_list = list()
        data = SolrInterface.get_simple_data_blob(query, 0, 1, 'ingest_date desc')
        peek_scan_size = requests.post(self.solr_cal_core_url,
                                       headers={'content-type': _APP_X_WWW_FORM_URLENCODED},
                                       timeout=5,
                                       data=data)
        if peek_scan_size.status_code == 200:
            single_response = peek_scan_size.json()
            data = SolrInterface.get_simple_data_blob(query, 0,
                                                      single_response['response']['numFound'],
                                                      'ingest_date desc')
            solr_rspns = requests.post(self.solr_cal_core_url,
                                       headers={'content-type': _APP_X_WWW_FORM_URLENCODED},
                                       timeout=5,
                                       data=data)
            for ncal in solr_rspns.json()['response']['docs']:
                ret_cal = self.build_basic_cal_dict(ncal)
                ret_cal_list.append(ret_cal)

            return ret_cal_list

        # The server could be down 404 (json decode will fail)
        # or we could have a malformed query 501(?) (with a json serializable response)
        raise SolrException(
            _SOLR_QUERY_FAILED_WITH_CODE.substitute(error_msg=peek_scan_size.status_code))

    def get_exec_block_calibrations(self, lookup_id, alma=False):
        r"""Get the cals for a single eb or mous"""
        try:
            id_str = str(lookup_id)
            if alma:
                q_str = f'mousId: "{id_str}"'
            else:
                q_str = f'exec_id: "{id_str}"'
            return self.query_cals(q_str)

        except Exception as any_except:
            ex = _EXCEPTION_MSG.substitute(c=__name__, d=def_name(), code=any_except)
            raise Exception(ex)

    def get_paged_eb_calibrations(self, id_list, alma=False):
        r"""Get the cals for a list of ebs or mouses"""
        try:
            ret_cal_list = list()

            max_count = 50
            id_chunks = [id_list[i * max_count:(i + 1) * max_count] for i in
                         range((len(id_list) + max_count - 1) // max_count)]

            for chunk in id_chunks:
                if len(chunk) > 0:
                    if alma:
                        query = 'mousId:('
                        mouses = ' OR '.join([f'"{s}"' for s in chunk])
                        query += mouses + ')'
                    else:
                        query = 'exec_id:('
                        ebs = ' OR '.join([str(s) for s in chunk])
                        query += ebs + ')'
                    ret_cal = self.query_cals(query)
                    ret_cal_list.extend(ret_cal)

            return ret_cal_list

        except Exception as any_except:
            ex = _EXCEPTION_MSG.substitute(c=__name__, d=def_name(), code=any_except)
            raise Exception(ex)

    def get_paged_images(self, **params_in):
        r"""Get a paginated list of images based on the search parameters provided"""
        try:
            query_params = ArchiveSolrParameters(**params_in)
            q_obj = construct_images_query(query_params)
            one_shot = requests.post(self.solr_img_core_url,
                                     headers={'content-type': _APP_X_WWW_FORM_URLENCODED},
                                     timeout=5,
                                     data=self.get_solr_query_data_block(query_params, q_obj))

            if one_shot.status_code == 200:
                images = one_shot.json()
                img_list = []
                img_total = images['response']['numFound']
                for img in images['response']['docs']:
                    curr_img = self.build_basic_img_dict(img)

                    img_list.append(curr_img)

                return img_list, img_total

            # The server could be down 404 (json decode will fail)
            # or we could have a malformed query 501(?) (with a json serializable response)
            raise Exception(
                _SOLR_QUERY_FAILED_WITH_CODE.substitute(error_msg=one_shot.status_code))

        except Exception as any_except:
            ex = _EXCEPTION_MSG.substitute(c=__name__, d=def_name(), code=any_except)
            raise SolrException(ex)

    def get_mous_size(self, mous_id, n_ebs, original_q):
        r"""Get the accumulated size of this mous, based on the number of ebs under it that
        match the query parameters provided"""
        try:
            # We have to include the original query here, because it could have filtered a number of
            # ebs under this mous out of the return list, but if we don't include the original here
            # we might accidentally get those non-matching ebs and that would throw off the size.
            # We are also assuming that the original query and this one will return the same number
            # of results, so we'll use the n_ebs for the row count.
            q_str = f'{original_q} AND mousId: "{mous_id}"'
            data = SolrInterface.get_simple_data_blob(q_str, 0, n_ebs, None)
            data['fl'] = 'access_estsize'
            one_shot = requests.post(self.solr_eb_core_url,
                                     headers={'content-type': _APP_X_WWW_FORM_URLENCODED},
                                     timeout=5,
                                     data=data)

            if one_shot.status_code == 200:
                respns = one_shot.json()
                accum_size = 0
                for ebs in respns['response']['docs']:
                    accum_size = accum_size + ebs['access_estsize']
                return accum_size

            # The server could be down 404 (json decode will fail)
            # or we could have a malformed query 501(?) (with a json serializable response)
            raise Exception(
                _SOLR_QUERY_FAILED_WITH_CODE.substitute(error_msg=one_shot.status_code))

        except Exception as any_except:
            ex = _EXCEPTION_MSG.substitute(c=__name__, d=def_name(), code=any_except)
            raise SolrException(ex)


class ResponseMsg:
    r"""A bit of a superfluous class used to wrap exception we catch in the view responses."""
    SUCCESS = 0
    SOFT_ERROR = 1
    HARD_ERROR = 2

    def __init__(self, flag, code, msg):
        self.status = flag
        self.code = code
        self.message = msg

    def __repr__(self):
        base_repr(self)

    def as_dict(self):
        r"""Return 'this' ResponseMsg as a dict"""
        return dict(
            success=self.status,
            code=self.code,
            message=self.message
        )
